# (C) Copyright 2005-2025 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Tests for the String trait type.

"""
import unittest

from traits.api import HasTraits, String
from traits.testing.optional_dependencies import numpy, requires_numpy


class A(HasTraits):
    string = String


class TestString(unittest.TestCase):
    @requires_numpy
    def test_accepts_numpy_string(self):
        numpy_string = numpy.str_("this is a numpy string!")
        a = A()
        a.string = numpy_string
        self.assertEqual(a.string, numpy_string)
        self.assertIs(type(a.string), str)



# (C) Copyright 2005-2025 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests to help find out if we can do type-safe casting. """

# Standard library imports.
import unittest

# Enthought library imports.
from traits.adaptation.api import reset_global_adaptation_manager
from traits.api import (
    adapt,
    Adapter,
    HasTraits,
    Instance,
    Int,
    Interface,
    provides,
    register_factory,
)

# Local imports.
from traits.interface_checker import InterfaceError, check_implements

# Make sure implicit interface checking is turned off, so that we can make the
# checks explicitly:
from traits import has_traits


class InterfaceCheckerTestCase(unittest.TestCase):
    """ Tests to help find out if we can do type-safe casting. """

    ###########################################################################
    # 'TestCase' interface.
    ###########################################################################

    def setUp(self):
        """ Prepares the test fixture before each test method is called. """
        # Make sure implicit interface checking is turned off, so that we can
        # make the checks explicitly:
        self._old_check_interfaces = has_traits.CHECK_INTERFACES
        has_traits.CHECK_INTERFACES = 0

        reset_global_adaptation_manager()

    def tearDown(self):
        has_traits.CHECK_INTERFACES = self._old_check_interfaces

    ###########################################################################
    # Tests.
    ###########################################################################

    def test_non_traits_class(self):
        """ non-traits class """

        class IFoo(Interface):
            def foo(self):
                pass

        # A class that *does* implement the interface.
        @provides(IFoo)
        class Foo(object):
            def foo(self):
                pass

        # The checker will raise an exception if the class does not implement
        # the interface.
        check_implements(Foo, IFoo, 2)

    def test_single_interface(self):
        """ single interface """

        class IFoo(Interface):
            x = Int

        # A class that *does* implement the interface.
        @provides(IFoo)
        class Foo(HasTraits):

            x = Int

        # The checker will raise an exception if the class does not implement
        # the interface.
        check_implements(Foo, IFoo, 2)

    def test_single_interface_with_invalid_method_signature(self):
        """ single interface with invalid method signature """

        class IFoo(Interface):
            def foo(self):
                pass

        # A class that does *not* implement the interface.
        @provides(IFoo)
        class Foo(HasTraits):
            # Extra argument!
            def foo(self, x):
                pass

        self.assertRaises(InterfaceError, check_implements, Foo, IFoo, 2)

    def test_single_interface_with_missing_trait(self):
        """ single interface with missing trait """

        class IFoo(Interface):
            x = Int

        # A class that does *not* implement the interface.
        @provides(IFoo)
        class Foo(HasTraits):
            pass

        self.assertRaises(InterfaceError, check_implements, Foo, IFoo, 2)

    def test_single_interface_with_missing_method(self):
        """ single interface with missing method """

        class IFoo(Interface):
            def method(self):
                pass

        # A class that does *not* implement the interface.
        @provides(IFoo)
        class Foo(HasTraits):
            pass

        self.assertRaises(InterfaceError, check_implements, Foo, IFoo, 2)

    def test_multiple_interfaces(self):
        """ multiple interfaces """

        class IFoo(Interface):
            x = Int

        class IBar(Interface):
            y = Int

        class IBaz(Interface):
            z = Int

        # A class that *does* implement the interface.
        @provides(IFoo, IBar, IBaz)
        class Foo(HasTraits):
            x = Int
            y = Int
            z = Int

        # The checker will raise an exception if the class does not implement
        # the interface.
        check_implements(Foo, [IFoo, IBar, IBaz], 2)

    def test_multiple_interfaces_with_invalid_method_signature(self):
        """ multiple interfaces with invalid method signature """

        class IFoo(Interface):
            def foo(self):
                pass

        class IBar(Interface):
            def bar(self):
                pass

        class IBaz(Interface):
            def baz(self):
                pass

        # A class that does *not* implement the interface.
        @provides(IFoo, IBar, IBaz)
        class Foo(HasTraits):
            def foo(self):
                pass

            def bar(self):
                pass

            # Extra argument!
            def baz(self, x):
                pass

        self.assertRaises(
            InterfaceError, check_implements, Foo, [IFoo, IBar, IBaz], 2
        )

    def test_multiple_interfaces_with_missing_trait(self):
        """ multiple interfaces with missing trait """

        class IFoo(Interface):
            x = Int

        class IBar(Interface):
            y = Int

        class IBaz(Interface):
            z = Int

        # A class that does *not* implement the interface.
        @provides(IFoo, IBar, IBaz)
        class Foo(HasTraits):

            x = Int
            y = Int

        self.assertRaises(
            InterfaceError, check_implements, Foo, [IFoo, IBar, IBaz], 2
        )

    def test_multiple_interfaces_with_missing_method(self):
        """ multiple interfaces with missing method """

        class IFoo(Interface):
            def foo(self):
                pass

        class IBar(Interface):
            def bar(self):
                pass

        class IBaz(Interface):
            def baz(self):
                pass

        # A class that does *not* implement the interface.
        @provides(IFoo, IBar, IBaz)
        class Foo(HasTraits):
            def foo(self):
                pass

            def bar(self):
                pass

        self.assertRaises(
            InterfaceError, check_implements, Foo, [IFoo, IBar, IBaz], 2
        )

    def test_inherited_interfaces(self):
        """ inherited interfaces """

        class IFoo(Interface):
            x = Int

        class IBar(IFoo):
            y = Int

        class IBaz(IBar):
            z = Int

        # A class that *does* implement the interface.
        @provides(IBaz)
        class Foo(HasTraits):
            x = Int
            y = Int
            z = Int

        # The checker will raise an exception if the class does not implement
        # the interface.
        check_implements(Foo, IBaz, 2)

    def test_inherited_interfaces_with_invalid_method_signature(self):
        """ inherited with invalid method signature """

        class IFoo(Interface):
            def foo(self):
                pass

        class IBar(IFoo):
            def bar(self):
                pass

        class IBaz(IBar):
            def baz(self):
                pass

        # A class that does *not* implement the interface.
        @provides(IBaz)
        class Foo(HasTraits):
            def foo(self):
                pass

            def bar(self):
                pass

            # Extra argument!
            def baz(self, x):
                pass

        self.assertRaises(InterfaceError, check_implements, Foo, IBaz, 2)

    def test_inherited_interfaces_with_missing_trait(self):
        """ inherited interfaces with missing trait """

        class IFoo(Interface):
            x = Int

        class IBar(IFoo):
            y = Int

        class IBaz(IBar):
            z = Int

        # A class that does *not* implement the interface.
        @provides(IBaz)
        class Foo(HasTraits):

            x = Int
            y = Int

        self.assertRaises(InterfaceError, check_implements, Foo, IBaz, 2)

    def test_inherited_interfaces_with_missing_method(self):
        """ inherited interfaces with missing method """

        class IFoo(Interface):
            def foo(self):
                pass

        class IBar(IFoo):
            def bar(self):
                pass

        class IBaz(IBar):
            def baz(self):
                pass

        # A class that does *not* implement the interface.
        @provides(IBaz)
        class Foo(HasTraits):
            def foo(self):
                pass

            def bar(self):
                pass

        self.assertRaises(InterfaceError, check_implements, Foo, IBaz, 2)

    def test_subclasses_with_wrong_signature_methods(self):
        """ Subclasses with incorrect method signatures """

        class IFoo(Interface):
            def foo(self, argument):
                pass

        @provides(IFoo)
        class Foo(HasTraits):
            def foo(self, argument):
                pass

        class Bar(Foo):
            def foo(self):
                pass

        self.assertRaises(InterfaceError, check_implements, Bar, IFoo, 2)

    # Make sure interfaces and adaptation etc still work with the 'HasTraits'
    # version of 'Interface'!
    def test_instance(self):
        """ instance """

        class IFoo(Interface):
            pass

        @provides(IFoo)
        class Foo(HasTraits):
            pass

        class Bar(HasTraits):
            foo = Instance(IFoo)

        Bar(foo=Foo())

    def test_callable(self):
        """ callable """

        class IFoo(Interface):
            pass

        @provides(IFoo)
        class Foo(HasTraits):
            pass

        f = Foo()

        self.assertEqual(f, adapt(f, IFoo))

    def test_adaptation(self):
        """ adaptation """

        class IFoo(Interface):
            pass

        class Foo(HasTraits):
            pass

        @provides(IFoo)
        class FooToIFooAdapter(Adapter):
            pass

        register_factory(FooToIFooAdapter, Foo, IFoo)

        f = Foo()

        # Make sure adaptation works.
        i_foo = adapt(f, IFoo)

        self.assertIsNotNone(i_foo)
        self.assertEqual(FooToIFooAdapter, type(i_foo))



# (C) Copyright 2005-2025 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import (
    Any,
    cached_property,
    ComparisonMode,
    HasTraits,
    Property,
    Str,
)


class NoneCompare(HasTraits):
    bar = Any(comparison_mode=ComparisonMode.none)


class IdentityCompare(HasTraits):
    bar = Any(comparison_mode=ComparisonMode.identity)


class EqualityCompare(HasTraits):
    bar = Any(comparison_mode=ComparisonMode.equality)


class Foo(HasTraits):
    """
    Class implementing custom equality.
    """

    name = Str

    def __eq__(self, other):
        return self.name == other.name


class TestComparisonMode(unittest.TestCase):
    def setUp(self):
        self.a = Foo(name="a")
        self.same_as_a = Foo(name="a")
        self.different_from_a = Foo(name="not a")

    def bar_changed(self, object, trait, old, new):
        self.changed_object = object
        self.changed_trait = trait
        self.changed_old = old
        self.changed_new = new
        self.changed_count += 1

    def reset_change_tracker(self):
        self.changed_object = None
        self.changed_trait = None
        self.changed_old = None
        self.changed_new = None
        self.changed_count = 0

    def check_tracker(self, object, trait, old, new, count):
        self.assertEqual(count, self.changed_count)
        self.assertIs(object, self.changed_object)
        self.assertEqual(trait, self.changed_trait)
        self.assertIs(old, self.changed_old)
        self.assertIs(new, self.changed_new)

    def test_none_first_assignment(self):
        nc = NoneCompare()
        nc.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = nc.bar
        nc.bar = self.a
        self.check_tracker(nc, "bar", default_value, self.a, 1)

    def test_identity_first_assignment(self):
        ic = IdentityCompare()
        ic.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ic.bar
        ic.bar = self.a
        self.check_tracker(ic, "bar", default_value, self.a, 1)

    def test_equality_first_assignment(self):
        ec = EqualityCompare()
        ec.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ec.bar
        ec.bar = self.a
        self.check_tracker(ec, "bar", default_value, self.a, 1)

    def test_none_same_object(self):
        nc = NoneCompare()
        nc.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = nc.bar
        nc.bar = self.a
        self.check_tracker(nc, "bar", default_value, self.a, 1)

        nc.bar = self.a
        self.check_tracker(nc, "bar", self.a, self.a, 2)

    def test_identity_same_object(self):
        ic = IdentityCompare()
        ic.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ic.bar
        ic.bar = self.a
        self.check_tracker(ic, "bar", default_value, self.a, 1)

        ic.bar = self.a
        self.check_tracker(ic, "bar", default_value, self.a, 1)

    def test_equality_same_object(self):
        ec = EqualityCompare()
        ec.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ec.bar
        ec.bar = self.a
        self.check_tracker(ec, "bar", default_value, self.a, 1)

        ec.bar = self.a
        self.check_tracker(ec, "bar", default_value, self.a, 1)

    def test_none_different_object(self):
        nc = NoneCompare()
        nc.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = nc.bar
        nc.bar = self.a
        self.check_tracker(nc, "bar", default_value, self.a, 1)

        nc.bar = self.different_from_a
        self.check_tracker(nc, "bar", self.a, self.different_from_a, 2)

    def test_identity_different_object(self):
        ic = IdentityCompare()
        ic.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ic.bar
        ic.bar = self.a
        self.check_tracker(ic, "bar", default_value, self.a, 1)

        ic.bar = self.different_from_a
        self.check_tracker(ic, "bar", self.a, self.different_from_a, 2)

    def test_equality_different_object(self):
        ec = EqualityCompare()
        ec.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ec.bar
        ec.bar = self.a
        self.check_tracker(ec, "bar", default_value, self.a, 1)

        ec.bar = self.different_from_a
        self.check_tracker(ec, "bar", self.a, self.different_from_a, 2)

    def test_none_different_object_same_as(self):
        nc = NoneCompare()
        nc.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = nc.bar
        nc.bar = self.a
        self.check_tracker(nc, "bar", default_value, self.a, 1)

        nc.bar = self.same_as_a
        self.check_tracker(nc, "bar", self.a, self.same_as_a, 2)

    def test_identity_different_object_same_as(self):
        ic = IdentityCompare()
        ic.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ic.bar
        ic.bar = self.a
        self.check_tracker(ic, "bar", default_value, self.a, 1)

        ic.bar = self.same_as_a
        self.check_tracker(ic, "bar", self.a, self.same_as_a, 2)

    def test_equality_different_object_same_as(self):
        ec = EqualityCompare()
        ec.on_trait_change(self.bar_changed, "bar")

        self.reset_change_tracker()

        default_value = ec.bar
        ec.bar = self.a
        self.check_tracker(ec, "bar", default_value, self.a, 1)

        # Values of a and same_as_a are the same and should therefore not
        # be considered a change.
        ec.bar = self.same_as_a
        self.check_tracker(ec, "bar", default_value, self.a, 1)

    def test_comparison_mode_none_with_cached_property(self):
        # Even though the property is cached such that old value equals new
        # value, its change event is tied to the dependent.

        class Model(HasTraits):
            value = Property(depends_on="name")
            name = Str(comparison_mode=ComparisonMode.none)

            @cached_property
            def _get_value(self):
                return self.trait_names

        instance = Model()
        events = []
        instance.on_trait_change(lambda: events.append(None), "value")

        instance.name = "A"

        events.clear()

        # when
        instance.name = "A"

        # then
        self.assertEqual(len(events), 1)



# (C) Copyright 2005-2025 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Tests for dynamic notifiers with `dispatch='ui'`.

Dynamic notifiers created with the `dispatch='ui'` option dispatch event
notifications on the UI thread. The class handling the dispatch,
`FastUITraitChangeNotifyWrapper`, is a subclass of `TraitChangeNotifyWrapper`.
Most of the functionality of the class is thus already covered by the
`TestDynamicNotifiers` test case, and we only need to test that the
notification really occurs on the UI thread.

At present, `dispatch='ui'` and `dispatch='fast_ui'` have the same effect.
"""

import asyncio
import contextlib
import threading
import unittest

from traits.api import (
    Callable,
    Float,
    get_ui_handler,
    HasTraits,
    on_trait_change,
    set_ui_handler,
)
from traits import trait_notifiers


class CalledAsMethod(HasTraits):
    foo = Float


class CalledAsDecorator(HasTraits):
    foo = Float

    callback = Callable

    @on_trait_change("foo", dispatch="ui")
    def on_foo_change(self, obj, name, old, new):
        self.callback(obj, name, old, new)


def asyncio_ui_handler(event_loop):
    """
    Create a UI handler that dispatches to the asyncio event loop.
    """

    def ui_handler(handler, *args, **kwargs):
        """
        UI handler that dispatches to the asyncio event loop.
        """
        event_loop.call_soon_threadsafe(lambda: handler(*args, **kwargs))

    return ui_handler


@contextlib.contextmanager
def use_asyncio_ui_handler(event_loop):
    """
    Context manager that temporarily sets the UI handler to an asyncio handler.
    """
    old_handler = get_ui_handler()
    set_ui_handler(asyncio_ui_handler(event_loop))
    try:
        yield
    finally:
        set_ui_handler(old_handler)


@contextlib.contextmanager
def clear_ui_handler():
    """
    Context manager that temporarily clears the UI handler.
    """
    old_handler = get_ui_handler()
    set_ui_handler(None)
    try:
        yield
    finally:
        set_ui_handler(old_handler)


class BaseTestUINotifiers(object):
    """Tests for dynamic notifiers with `dispatch='ui'`."""

    #### 'TestCase' protocol ##################################################

    def setUp(self):
        self.notifications = []
        self.exceptions = []
        self.done = asyncio.Event()
        self.obj = self.obj_factory()

    def enterContext(self, cm):
        # Backport of Python 3.11's TestCase.enterContext method.
        result = type(cm).__enter__(cm)
        self.addCleanup(type(cm).__exit__, cm, None, None, None)
        return result

    #### 'TestUINotifiers' protocol ###########################################

    def modify_obj(self):
        trait_notifiers.push_exception_handler(
            lambda *args: None, reraise_exceptions=True
        )
        try:
            self.obj.foo = 3
        except Exception as e:
            self.exceptions.append(e)
        finally:
            trait_notifiers.pop_exception_handler()

    def on_foo_notifications(self, obj, name, old, new):
        event = (threading.current_thread(), (obj, name, old, new))
        self.notifications.append(event)
        self.done.set()

    #### Tests ################################################################

    def test_notification_from_main_thread_with_no_ui_handler(self):
        # Given
        self.enterContext(clear_ui_handler())

        # When we set obj.foo to 3 on the main thread.
        self.modify_obj()

        # Then the notification is processed synchronously on the main thread.
        self.assertEqual(
            self.notifications,
            [(threading.main_thread(), (self.obj, "foo", 0, 3))],
        )

    def test_notification_from_main_thread_with_registered_ui_handler(self):
        # Given
        self.enterContext(use_asyncio_ui_handler(asyncio.get_event_loop()))

        # When we set obj.foo to 3 on the main thread.
        self.modify_obj()

        # Then the notification is processed synchronously on the main thread.
        self.assertEqual(
            self.notifications,
            [(threading.main_thread(), (self.obj, "foo", 0, 3))],
        )

    def test_notification_from_separate_thread_failure_case(self):
        # Given no registered ui handler
        self.enterContext(clear_ui_handler())

        # When we set obj.foo to 3 on a separate thread, and wait for
        # that thread to complete.
        background_thread = threading.Thread(target=self.modify_obj)
        background_thread.start()
        background_thread.join()

        # Then no notification is processed ...
        self.assertEqual(self.notifications, [])

        # ... and an error was raised
        self.assertEqual(len(self.exceptions), 1)
        self.assertIsInstance(self.exceptions[0], RuntimeError)
        self.assertIn("no UI handler registered", str(self.exceptions[0]))

        # ... but the attribute change was still applied.
        self.assertEqual(self.obj.foo, 3)

    async def test_notification_from_separate_thread(self):
        # Given an asyncio ui handler
        self.enterContext(use_asyncio_ui_handler(asyncio.get_event_loop()))

        # When we set obj.foo to 3 on a separate thread.
        background_thread = threading.Thread(target=self.modify_obj)
        background_thread.start()
        self.addCleanup(background_thread.join)

        # Then the notification will eventually be processed on the main
        # thread.
        await asyncio.wait_for(self.done.wait(), timeout=5.0)
        self.assertEqual(
            self.notifications,
            [(threading.main_thread(), (self.obj, "foo", 0, 3))],
        )
        self.assertEqual(self.obj.foo, 3)


class TestMethodUINotifiers(
    BaseTestUINotifiers, unittest.IsolatedAsyncioTestCase
):
    """Tests for dynamic notifiers with `dispatch='ui'` set by method call."""

    def obj_factory(self):
        obj = CalledAsMethod()
        obj.on_trait_change(self.on_foo_notifications, "foo", dispatch="ui")
        return obj


class TestDecoratorUINotifiers(
    BaseTestUINotifiers, unittest.IsolatedAsyncioTestCase
):
    """Tests for dynamic notifiers with `dispatch='ui'` set by decorator."""

    def obj_factory(self):
        return CalledAsDecorator(callback=self.on_foo_notifications)



# (C) Copyright 2005-2025 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, Int


class TestTraitNotifiers(unittest.TestCase):
    def test_ui_dispatch(self):
        # Given
        class DispatchTest(HasTraits):
            test_param = Int()

        t = DispatchTest()

        event_list = []

        # create handler
        def test_handler(event):
            event_list.append(event)

        # When
        t.observe(test_handler, "test_param", dispatch="ui")
        t.test_param = 1

        # Then
        # check the observer is called once
        self.assertEqual(len(event_list), 1)
        # check the name of the parameter change
        self.assertEqual(event_list[0].name, "test_param")
        # check whether the value starts at 0
        self.assertEqual(event_list[0].old, 0)
        # check whether the value has been set to 1
        self.assertEqual(event_list[0].new, 1)



# (C) Copyright 2005-2025 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, Int, List, Range, Str, TraitError, Tuple


class WithFloatRange(HasTraits):
    r = Range(0.0, 100.0)
    r_copied_on_change = Str

    _changed_handler_calls = Int

    def _r_changed(self, old, new):
        self._changed_handler_calls += 1
        self.r_copied_on_change = str(self.r)

        if (self.r % 10) > 0:
            self.r += 10 - (self.r % 10)


class WithLargeIntRange(HasTraits):
    r = Range(0, 1000)
    r_copied_on_change = Str

    _changed_handler_calls = Int

    def _r_changed(self, old, new):
        self._changed_handler_calls += 1
        self.r_copied_on_change = str(self.r)

        if self.r > 100:
            self.r = 0


class WithDynamicRange(HasTraits):
    low = Int(0)
    high = Int(10)
    value = Int(3)

    r = Range(value="value", low="low", high="high", exclude_high=True)

    def _r_changed(self, old, new):
        self._changed_handler_calls += 1


class RangeTestCase(unittest.TestCase):
    def test_non_ui_events(self):

        obj = WithFloatRange()
        obj._changed_handler_calls = 0

        obj.r = 10
        self.assertEqual(1, obj._changed_handler_calls)

        obj._changed_handler_calls = 0
        obj.r = 34.56
        self.assertEqual(obj._changed_handler_calls, 2)
        self.assertEqual(obj.r, 40)

    def test_non_ui_int_events(self):

        # Even though the range is configured for 0..1000, the handler resets
        # the value to 0 when it exceeds 100.
        obj = WithLargeIntRange()
        obj._changed_handler_calls = 0

        obj.r = 10
        self.assertEqual(obj._changed_handler_calls, 1)
        self.assertEqual(obj.r, 10)

        obj.r = 100
        self.assertEqual(obj._changed_handler_calls, 2)
        self.assertEqual(obj.r, 100)

        obj.r = 101
        self.assertEqual(obj._changed_handler_calls, 4)
        self.assertEqual(obj.r, 0)

    def test_dynamic_events(self):

        obj = WithDynamicRange()
        obj._changed_handler_calls = 0

        obj.r = 5
        self.assertEqual(obj._changed_handler_calls, 1)
        self.assertEqual(obj.r, 5)

        with self.assertRaises(TraitError):
            obj.r = obj.high
        self.assertEqual(obj.r, 5)

    def test_dynamic_range_in_tuple(self):
        # Regression test for #1391
        class HasRangeInTuple(HasTraits):
            low = Int()
            high = Int()
            hours_and_name = Tuple(Range(low="low", high="high"), Str)

        model = HasRangeInTuple(low=0, high=48)
        model.hours_and_name = (40, "fred")
        self.assertEqual(model.hours_and_name, (40, "fred"))
        with self.assertRaises(TraitError):
            # First argument out or range; should raise.
            model.hours_and_name = (50, "george")

    def test_dynamic_range_in_list(self):
        # Another regression test for #1391.
        class HasRangeInList(HasTraits):
            #: Valid digit range
            low = Int()

            high = Int()

            #: Sequence of digits
            digit_sequence = List(Range(low="low", high="high"))

        model = HasRangeInList(low=-1, high=1)
        model.digit_sequence = [-1, 0, 1, 1]
        with self.assertRaises(TraitError):
            model.digit_sequence = [-1, 0, 2, 1]
